@@ -6,10 +6,8 @@
 # SPDX-License-Identifier: BSD-2-Clause
 
 from migen import *
-from migen.genlib.misc import chooser, displacer
-from migen.genlib.fifo import SyncFIFOBuffered
 
-from litex.soc.interconnect.stream import *
+from litex.soc.interconnect import stream
 from litex.soc.interconnect.csr import *
 
 from litepcie.common import *
@@ -17,11 +15,8 @@ from litepcie.tlp.common import *
 
 # Constants/Layouts --------------------------------------------------------------------------------
 
-DMA_IRQ_DISABLE    = 0
-DMA_LAST_DISABLE   = 1
-
 def descriptor_layout(with_user_id=False):
-    layout = [("address", 32), ("length",  24), ("control", 8)]
+    layout = [("address", 32), ("length",  24), ("irq_disable", 1), ("last_disable", 1)]
     if with_user_id:
         layout += [("user_id", 8)]
     return EndpointDescription(layout)
@@ -32,31 +27,48 @@ def descriptor_layout(with_user_id=False):
 class LitePCIeDMAScatterGather(Module, AutoCSR):
     """LitePCIe DMA Scatter-Gather
 
-    Software programmable table that stores a list of DMA descriptors to be executed.
+    Software programmable table storing a list of DMA descriptors.
+
+                               Mode
+                                │
+                              ┌─▼─┐  ┌───────────────────┐
+       Descriptor to program  │   │  │                   │
+             (Prog Mode) ─────►   │  │                   │
+                              │   │  │    ScatterGather  │
+                              │   ├──►     Table(FIFO)   ├──┬───► Descriptor (To DMA).
+                          ┌───►   │  │                   │  │
+                          │   │   │  │                   │  │
+                          │   └───┘  └───────────────────┘  │
+                          │                                 │
+                          └─────────────────────────────────┘
+                                    Refill (Loop Mode)
+
 
     A DMA descriptor is composed of:
-    - a 32-bit address: the base address where the data stream should be written/read.
-    - a 24-bit length : the length of the data stream (bytes).
-    - 8 control bits  : control to allow dynamic specific behavior (for example: disable MSI IRQ).
+    - a 32-bit address: The base address of the Host where the data stream should be written/read.
+    - a 24-bit length : The length of the data stream (bytes).
+    - a 8-bit control : Dynamic controls (ex: Disable IRQ generation, disable Last handling).
 
     The table is implemented as a FIFO initially filled by software. Once enabled, the DMA gets the
     descriptors from this table and executes them.
 
     This module has two modes:
-    - PROG mode: Used to program the table by software and for cases where automatic refill of the
+    - Prog mode: Used to program the table by software and for cases where automatic refill of the
     table is not needed: A descriptor is only executed once and when all the descriptors have been
     executed (ie the table is empty), the DMA just stops until the next software refill.
-    - LOOP mode: Used once the table has been filled by software in PROG mode and allow continuous
+    - Loop mode: Used once the table has been filled by software in PROG mode and allow continuous
     Scatter-Gather DMA: Each descriptor sent to the DMA is refilled to the table.
 
-    In LOOP mode, a loop status is maintained by the hardware for the software synchronization of the
+    In Loop mode, a loop status is maintained by the hardware for the software synchronization of the
     DMA buffers. (Even if a MSI IRQ is generated after a descriptor has been executed, since IRQ can
     potentially be lost, it's safer for the software to just use the hardware loop status than to
     maintain a software loop status based MSI IRQ reception).
     """
     def __init__(self, depth):
+        # Stream Endpoint.
         self.source = source = stream.Endpoint(descriptor_layout())
 
+        # Control/Status.
         self.value = CSRStorage(64, reset_less=True, fields=[
             CSRField("address",      size=32, description="32-bit Address of the descriptor (bytes-aligned)."),
             CSRField("length",       size=24, description="24-bit Length  of the descriptor (in bytes)."),
@@ -65,12 +77,12 @@ class LitePCIeDMAScatterGather(Module, AutoCSR):
             ], description="64-bit DMA descriptor to be written to the table.")
         self.we = CSRStorage(description="A write to this register adds the descriptor to table.")
         self.loop_prog_n = CSRStorage(description="""Mode Selection.\n
-            ``0``: **PROG** mode / ``1``: **LOOP** mode.\n
-            **PROG** mode should be used to program the table by software and for cases where automatic
+            ``0``: **Prog** mode / ``1``: **Loop** mode.\n
+            **Prog** mode should be used to program the table by software and for cases where automatic
             refill of the table is not needed: A descriptor is only executed once and when all the
             descriptors have been executed (ie the table is empty), the DMA just stops until the next
             software refill.\n
-            **LOOP** mode should be used once the table has been filled by software in **PROG** mode
+            **Loop** mode should be used once the table has been filled by software in **Prog** mode
             and allow continuous Scatter-Gather DMA: Each descriptor sent to the DMA is refilled to the table.
             """)
         self.loop_status = CSRStatus(fields=[
@@ -78,79 +90,54 @@ class LitePCIeDMAScatterGather(Module, AutoCSR):
             CSRField("count", size=16, description= "Loops of the DMA descriptor table since started."),
             ], description="Loop monitoring for software synchronization.")
         self.level = CSRStatus(bits_for(depth), description="Number descriptors in the table.")
-        self.flush = CSRStorage(description="A write to this register flushes the table.")
+        self.reset = CSRStorage(description="A write to this register resets the table.")
 
         # # #
 
-        # CSRs -------------------------------------------------------------------------------------
-        address     = self.value.fields.address
-        length      = self.value.fields.length
-        control     = Cat(self.value.fields.irq_disable, self.value.fields.last_disable)
-        we          = self.we.storage & self.we.re
-        loop_prog_n = self.loop_prog_n.storage
-        loop_status = self.loop_status
-        level       = self.level.status
-        flush       = self.flush.storage & self.flush.re
+        # Table (FIFO) -----------------------------------------------------------------------------
+        table = stream.SyncFIFO(descriptor_layout(), depth)
+        table = ResetInserter()(table)
+        self.submodules += table
+        self.comb += table.reset.eq(self.reset.storage & self.reset.re)
+        self.comb += self.level.status.eq(table.level)
 
-        # Table FIFO -------------------------------------------------------------------------------
-        fifo = SyncFIFO([("address", 32), ("length",  24), ("control", 8)], depth)
-        fifo = ResetInserter()(fifo)
-        self.submodules += fifo
-        self.comb += [
-            fifo.reset.eq(flush),
-            level.eq(fifo.level),
-        ]
-
-        # Write logic ------------------------------------------------------------------------------
+        # Table Write logic ------------------------------------------------------------------------
+        prog_mode = (self.loop_prog_n.storage == 0)
         self.sync += [
-            # In LOOP mode, the FIFO is automatically refilled with the descriptors.
-            If(loop_prog_n,
-                fifo.sink.address.eq(fifo.source.address),
-                fifo.sink.length.eq(fifo.source.length),
-                fifo.sink.control.eq(fifo.source.control),
-                fifo.sink.first.eq(fifo.source.first),
-                fifo.sink.valid.eq(fifo.source.ready)
-            # In PROG mode, the FIFO is filled through the CSRs and not automatically refilled.
+            # In Prog mode, the Table is filled through the CSRs.
+            If(prog_mode,
+                table.sink.address.eq(self.value.fields.address),
+                table.sink.length.eq(self.value.fields.length),
+                table.sink.irq_disable.eq(self.value.fields.irq_disable),
+                table.sink.last_disable.eq(self.value.fields.last_disable),
+                table.sink.first.eq(table.level == 0),
+                table.sink.valid.eq(self.we.storage & self.we.re),
+            # In Loop mode, the Table is automatically refilled.
             ).Else(
-                fifo.sink.address.eq(address),
-                fifo.sink.length.eq(length),
-                fifo.sink.control.eq(control),
-                fifo.sink.first.eq(~fifo.source.valid),
-                fifo.sink.valid.eq(we)
+                table.source.connect(table.sink, omit={"valid", "ready"}),
+                table.sink.valid.eq(table.source.valid & table.source.ready),
             )
         ]
 
-        # Read logic -------------------------------------------------------------------------------
-        self.comb += [
-            source.valid.eq(fifo.source.valid),
-            source.first.eq(fifo.source.first),
-            fifo.source.ready.eq(source.valid & source.ready),
-            source.address.eq(fifo.source.address),
-            source.length.eq(fifo.source.length),
-            source.control.eq(fifo.source.control),
-        ]
+        # Table Read logic -------------------------------------------------------------------------
+        self.comb += table.source.connect(source)
 
-        # Loops monitoring -------------------------------------------------------------------------
-        # Used for software synchronization in LOOP mode.
-        loop_first = Signal(reset=1)
-        loop_index = Signal(log2_int(depth))
-        loop_count = Signal(16)
+        # Loop Status (For Software Sychronization in Loop mode) -----------------------------------
+        loop_index = self.loop_status.fields.index
+        loop_count = self.loop_status.fields.count
         self.sync += [
-            If(flush,
-                loop_first.eq(1),
+            # Reset Loop Index/Count on Table reset.
+            If(table.reset,
+                # Init count to 2**n-1 to have (index, count) == (0, 0) for the first Descriptor.
                 loop_index.eq(0),
-                loop_count.eq(0),
-                loop_status.fields.index.eq(0),
-                loop_status.fields.count.eq(0),
-            ).Elif(source.valid & source.ready,
-                loop_status.fields.index.eq(loop_index),
-                loop_status.fields.count.eq(loop_count),
-                If(source.first,
-                    loop_first.eq(0),
+                loop_count.eq(2**len(loop_count) - 1),
+            # When a Descriptor is consumned...
+            ).Elif(table.source.valid & table.source.ready,
+                # Update Loop Status with current Loop Index/Count.
+                # Update Loop Index/Count.
+                If(table.source.first,
                     loop_index.eq(0),
-                    If(~loop_first,
-                        loop_count.eq(loop_count + 1)
-                    )
+                    loop_count.eq(loop_count + 1),
                 ).Else(
                     loop_index.eq(loop_index + 1)
                 )
@@ -173,55 +160,67 @@ class LitePCIeDMADescriptorSplitter(Module, AutoCSR):
     several shorter descriptors.
     """
     def __init__(self, max_size):
+        # Stream Endpoints.
         self.sink   =   sink = stream.Endpoint(descriptor_layout())
         self.source = source = stream.Endpoint(descriptor_layout(with_user_id=True))
+
+        # Control.
         self.end    = Signal()
 
         # # #
 
-        offset  = Signal(32)
-        user_id = Signal(32)
+        desc_length  = Signal(32)
+        desc_offset  = Signal(32)
+        desc_id      = Signal(32)
 
         # FSM --------------------------------------------------------------------------------------
         self.submodules.fsm = fsm = FSM(reset_state="IDLE")
         fsm.act("IDLE",
-            NextValue(offset, 0),
+            # Set Descriptor Offset/Length.
+            NextValue(desc_offset, 0),
+            NextValue(desc_length, sink.length),
+            # Wait for a Descriptor and go to Run.
             If(sink.valid,
-                NextState("RUN")
-            ).Else(
-                sink.ready.eq(1)
+                NextState("SPLIT")
             )
         )
+        # Split Data-Path.
         self.comb += [
-            source.address.eq(sink.address + offset),
-            source.control.eq(sink.control),
-            source.user_id.eq(user_id),
+            source.address.eq(sink.address + desc_offset),
+            source.irq_disable.eq(sink.irq_disable),
+            source.last_disable.eq(sink.last_disable),
+            source.user_id.eq(desc_id),
         ]
-        fsm.act("RUN",
+        fsm.act("SPLIT",
+            # Split Control-Path.
             source.valid.eq(1),
-            source.first.eq(offset == 0),
-            If((sink.length - offset) > max_size,
+            source.first.eq(desc_offset == 0),
+            # Full Descriptor when Length > max_size.
+            If(desc_length > max_size,
                 source.last.eq(self.end),
                 source.length.eq(max_size),
-                If(source.ready,
-                    NextValue(offset, offset + max_size),
-                    If(self.end,
-                        NextState("ACK")
-                    )
-                )
+            # Partial Descriptor when Length <= max_size
             ).Else(
                 source.last.eq(1),
-                source.length.eq(sink.length - offset),
-                If(source.ready,
-                    NextState("ACK")
+                source.length.eq(desc_length),
+            ),
+            # When Descriptor is accepted...
+            If(source.ready,
+                # Increment Offset.
+                NextValue(desc_offset, desc_offset + max_size),
+                # Decrement Length.
+                NextValue(desc_length, desc_length - max_size),
+                # When Last....
+                If(source.last,
+                    # Accept Descriptor.
+                    sink.ready.eq(1),
+                    # Increment ID.
+                    NextValue(desc_id, desc_id + 1),
+                    # Go to Idle.
+                    NextState("IDLE")
                 )
             )
         )
-        fsm.act("ACK",
-            sink.ready.eq(1),
-            NextValue(user_id, user_id + 1),
-            NextState("IDLE")
-        )
 
 # LitePCIeDMAReader --------------------------------------------------------------------------------
 
@@ -240,23 +239,24 @@ class LitePCIeDMAReader(Module, AutoCSR):
     A MSI IRQ can be generated when a descriptor has been executed.
     """
     def __init__(self, endpoint, port, with_table=True, table_depth=256):
-        self.port   = port
+        self.port = port
+        # Stream Endpoint.
         self.source = stream.Endpoint(dma_layout(endpoint.phy.data_width))
-        self.irq    = Signal()
+
+        # Control.
         self.enable = CSRStorage(description="DMA Reader Control. Write ``1`` to enable DMA Reader.", reset=0 if with_table else 1)
 
+        # IRQ.
+        self.irq = Signal()
+
         # # #
 
         # CSR/Parameters ---------------------------------------------------------------------------
         enable = self.enable.storage
 
+        length_shift          = log2_int(endpoint.phy.data_width//8)
         max_words_per_request = max_request_size//(endpoint.phy.data_width//8)
         max_pending_words     = endpoint.max_pending_requests*max_words_per_request
-        fifo_depth            = 4*max_pending_words
-
-        pending_words         = Signal(max=fifo_depth + 1)
-        pending_words_queue   = Signal.like(pending_words)
-        pending_words_dequeue = Signal.like(pending_words)
 
         # Table ------------------------------------------------------------------------------------
         if with_table:
@@ -268,48 +268,67 @@ class LitePCIeDMAReader(Module, AutoCSR):
         # DMA descriptors need to be splitted in descriptors of max_request_size (negotiated at link-up)
         splitter = LitePCIeDMADescriptorSplitter(max_size=endpoint.phy.max_request_size)
         splitter = ResetInserter()(splitter)
-        splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter)
+        splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter) # For timings.
         self.submodules.splitter = splitter
         if with_table:
             self.comb += self.table.source.connect(splitter.sink)
         else:
             self.comb += self.desc_sink.connect(splitter.sink)
 
-        # Data FIFO --------------------------------------------------------------------------------
-        fifo = SyncFIFO(dma_layout(endpoint.phy.data_width), fifo_depth, buffered=True)
-        fifo = ResetInserter()(fifo)
-        self.submodules.fifo = fifo
-        self.comb += fifo.source.connect(self.source)
-
+        # User ID ----------------------------------------------------------------------------------
         last_user_id = Signal(8, reset=255)
+        self.sync += If(port.sink.valid & port.sink.first & port.sink.ready,
+            last_user_id.eq(port.sink.user_id)
+        )
+
+        # Data FIFO --------------------------------------------------------------------------------
+        data_fifo_depth = 4*max_pending_words
+        data_fifo = SyncFIFO(dma_layout(endpoint.phy.data_width), data_fifo_depth, buffered=True)
+        self.submodules += ResetInserter()(data_fifo)
         self.comb += [
-            fifo.sink.valid.eq(port.sink.valid),
-            fifo.sink.first.eq(port.sink.first & (port.sink.user_id != last_user_id)),
-            fifo.sink.data.eq(port.sink.dat),
-            port.sink.ready.eq(fifo.sink.ready | ~enable),
-        ]
-        self.sync += [
-            If(port.sink.valid & port.sink.first & port.sink.ready,
-                last_user_id.eq(port.sink.user_id)
+            # Connect Data FIFO to Source.
+            data_fifo.source.connect(self.source),
+            # When Enabled, connect Sink to Data FIFO.
+            If(enable,
+                port.sink.connect(data_fifo.sink, keep={"valid", "ready"}),
+                data_fifo.sink.data.eq(port.sink.dat),
+                data_fifo.sink.first.eq(port.sink.first & (port.sink.user_id != last_user_id)),
+            # Else accept incoming Port Data.
+            ).Else(
+                port.sink.ready.eq(1)
             )
         ]
 
+        # Pending words ----------------------------------------------------------------------------
+        pending_words         = Signal(max=data_fifo_depth + 1)
+        pending_words_queue   = Signal.like(pending_words)
+        pending_words_dequeue = Signal.like(pending_words)
+        self.comb += [
+            # Queue Pending words as Read Requests are emitted.
+            If(splitter.source.valid & splitter.source.ready,
+                pending_words_queue.eq(splitter.source.length[length_shift:])
+            ),
+            # Dequeue Pending words as Read Responses are received.
+            If(data_fifo.source.valid & data_fifo.source.ready,
+                pending_words_dequeue.eq(1)
+            ),
+        ]
+        # Update Pending words.
+        self.sync += pending_words.eq(pending_words + pending_words_queue - pending_words_dequeue)
+        self.sync += If(~enable, pending_words.eq(0))
+
         # FSM --------------------------------------------------------------------------------------
         self.submodules.fsm = fsm = FSM(reset_state="IDLE")
         fsm.act("IDLE",
-            # Reset Splitter/FIFO when DMA is disabled.
-            If(~enable,
-                splitter.reset.eq(1),
-                fifo.reset.eq(1)
-            ),
-            # Wait for a Request from the Splitter.
-            If(splitter.source.valid,
-                # When enough space in the FIFO, generate the Request.
-                If(pending_words < (fifo_depth - max_words_per_request),
-                    NextState("REQUEST"),
-                )
+            # Reset Splitter/FIFO when Disabled.
+            splitter.reset.eq( ~enable),
+            data_fifo.reset.eq(~enable),
+            # Wait for a Descriptor and to have enough Space to generate the Request.
+            If(splitter.source.valid & (pending_words < (data_fifo_depth - max_words_per_request)),
+                NextState("REQUEST"),
             )
         )
+        # Request Data-Path.
         self.comb += [
             port.source.channel.eq(port.channel),
             port.source.user_id.eq(splitter.source.user_id),
@@ -322,36 +341,20 @@ class LitePCIeDMAReader(Module, AutoCSR):
             port.source.dat.eq(0),
         ]
         fsm.act("REQUEST",
-            # Generate Request.
+            # Request Control-Path.
             port.source.valid.eq(1),
+            # When Request is accepted...
             If(port.source.ready,
-                # Ack Splitter when Request is accepted.
+                # Accept Descriptor.
                 splitter.source.ready.eq(1),
+                # Return to Idle.
                 NextState("IDLE"),
             )
         )
 
-        # Pending words ----------------------------------------------------------------------------
-        self.comb += [
-            # Queue Pending words as Read Requests are emitted.
-            If(splitter.source.valid & splitter.source.ready,
-                pending_words_queue.eq(splitter.source.length[log2_int(endpoint.phy.data_width//8):])
-            ),
-            # Dequeue Pending words as Read Responses are received.
-            If(fifo.source.valid & fifo.source.ready,
-                pending_words_dequeue.eq(1)
-            ),
-        ]
-        # Update Pending words.
-        self.sync += pending_words.eq(pending_words + pending_words_queue - pending_words_dequeue)
-        self.sync += If(~enable, pending_words.eq(0))
-
         # IRQ --------------------------------------------------------------------------------------
-        self.comb += self.irq.eq(
-            splitter.source.valid &
-            splitter.source.ready &
-            splitter.source.last  &
-            ~splitter.source.control[DMA_IRQ_DISABLE]
+        self.comb += If(splitter.source.valid & splitter.source.ready & splitter.source.last,
+            self.irq.eq(~splitter.source.irq_disable)
         )
 
 # LitePCIeDMAWriter --------------------------------------------------------------------------------
@@ -371,20 +374,23 @@ class LitePCIeDMAWriter(Module, AutoCSR):
     A MSI IRQ can be generated when a descriptor has been executed.
     """
     def __init__(self, endpoint, port, with_table=True, table_depth=256):
-        self.port   = port
+        self.port = port
+        # Stream Endpoint.
         self.sink   = sink = stream.Endpoint(dma_layout(endpoint.phy.data_width))
-        self.irq    = Signal()
+
+        # Control.
         self.enable = CSRStorage(description="DMA Writer Control. Write ``1`` to enable DMA Writer.", reset=0 if with_table else 1)
 
-        # # #
+        # IRQ.
+        self.irq = Signal()
 
-        counter = Signal(max=(2**len(endpoint.phy.max_payload_size))//8)
+        # # #
 
         # CSR/Parameters ---------------------------------------------------------------------------
         enable = self.enable.storage
 
+        length_shift          = log2_int(endpoint.phy.data_width//8)
         max_words_per_request = max_payload_size//(endpoint.phy.data_width//8)
-        fifo_depth            = 4*max_words_per_request
 
         # Table ------------------------------------------------------------------------------------
         if with_table:
@@ -396,7 +402,7 @@ class LitePCIeDMAWriter(Module, AutoCSR):
         # DMA descriptors need to be splitted in descriptors of max_request_size (negotiated at link-up)
         splitter = LitePCIeDMADescriptorSplitter(max_size=endpoint.phy.max_payload_size)
         splitter = ResetInserter()(splitter)
-        splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter)
+        splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter) # For timings.
         self.submodules.splitter = splitter
         if with_table:
             self.comb += self.table.source.connect(splitter.sink)
@@ -404,67 +410,70 @@ class LitePCIeDMAWriter(Module, AutoCSR):
             self.comb += self.desc_sink.connect(splitter.sink)
 
         # Data FIFO --------------------------------------------------------------------------------
-        fifo = SyncFIFOBuffered(endpoint.phy.data_width + 1, fifo_depth)
-        self.submodules += ResetInserter()(fifo)
+        data_fifo_depth = 4*max_words_per_request
+        data_fifo = stream.SyncFIFO([("data", endpoint.phy.data_width)], data_fifo_depth, buffered=True)
+        self.submodules += ResetInserter()(data_fifo)
         self.comb += [
-            fifo.we.eq(sink.valid & enable),
-            sink.ready.eq(fifo.writable | ~enable),
-            fifo.din.eq(Cat(sink.data, sink.last))
+            # When Enabled, connect Sink to Data FIFO.
+            If(enable,
+                sink.connect(data_fifo.sink)
+            # Else accept incoming Sink Data.
+            ).Else(
+                sink.ready.eq(1)
+            )
         ]
 
         # FSM --------------------------------------------------------------------------------------
+        req_count = Signal.like(splitter.source.length)
         self.submodules.fsm = fsm = FSM(reset_state="IDLE")
         fsm.act("IDLE",
-            # Reset Splitter/FIFO when DMA is disabled.
-            If(~enable,
-                splitter.reset.eq(1),
-                fifo.reset.eq(1)
-            ),
-            NextValue(counter, 0),
-             # Wait for a Request from the Splitter.
-            If(splitter.source.valid,
-                # When enough Data in the FIFO, generate the Request.
-                If(fifo.level >= splitter.source.length[log2_int(endpoint.phy.data_width//8):],
-                    NextState("REQUEST"),
-                )
+            # Reset Splitter/FIFO when Disabled.
+            splitter.reset.eq( ~enable),
+            data_fifo.reset.eq(~enable),
+            # Reset Request Count.
+            NextValue(req_count, 0),
+            # Wait for a Descriptor and to have enough Data to generate the Request.
+            If(splitter.source.valid & (data_fifo.level >= splitter.source.length[length_shift:]),
+                NextState("REQUEST"),
             )
         )
-        length_shift = log2_int(endpoint.phy.data_width//8)
+        # Request Data-Path.
         self.comb += [
             port.source.channel.eq(port.channel),
             port.source.user_id.eq(splitter.source.user_id),
-            port.source.first.eq(counter == 0),
-            port.source.last.eq((counter == splitter.source.length[length_shift:] - 1)),
+            port.source.first.eq(req_count == 0),
+            port.source.last.eq((req_count == splitter.source.length[length_shift:] - 1)),
             port.source.we.eq(1),
             port.source.adr.eq(splitter.source.address),
             port.source.req_id.eq(endpoint.phy.id),
             port.source.tag.eq(0),
             port.source.len.eq(splitter.source.length[2:]),
-            port.source.dat.eq(fifo.dout[:-1])
+            port.source.dat.eq(data_fifo.source.data)
         ]
         fsm.act("REQUEST",
-            # Generate the Request.
+            # Request Control-Path.
             port.source.valid.eq(1),
+            # When Request is accepted...
             If(port.source.ready,
-                NextValue(counter, counter + 1),
-                # Read only if not last.
-                fifo.re.eq(~(fifo.dout[-1] & ~splitter.source.control[DMA_LAST_DISABLE])),
+                # Increment Request Count.
+                NextValue(req_count, req_count + 1),
+                # Accept Data.
+                data_fifo.source.ready.eq(1),
+                # When last...
                 If(port.source.last,
-                    # Always read.
-                    fifo.re.eq(1),
-                    splitter.end.eq(fifo.dout[-1] & ~splitter.source.control[DMA_LAST_DISABLE]),
+                    # Accept Descriptor.
                     splitter.source.ready.eq(1),
+                    # Early Splitter End (can be disabled).
+                    splitter.end.eq(data_fifo.source.last & ~splitter.source.last_disable),
+                    # Return to Idle.
                     NextState("IDLE"),
                 )
             )
         )
 
         # IRQ --------------------------------------------------------------------------------------
-        self.comb += self.irq.eq(
-            splitter.source.valid &
-            splitter.source.ready &
-            splitter.source.last  &
-            ~splitter.source.control[DMA_IRQ_DISABLE]
+        self.comb += If(splitter.source.valid & splitter.source.ready & splitter.source.last,
+            self.irq.eq(~splitter.source.irq_disable)
         )
 
 # LitePCIeDMALoopback ------------------------------------------------------------------------------
@@ -564,42 +573,106 @@ class LitePCIeDMABuffering(Module, AutoCSR):
     appears in the streams and our Writes/Reads can't be absorbed/produced at a fixed rate. A minimum
     of buffering is needed to make sure the gaps are smoothed and not propagated to user modules.
     """
-    def __init__(self, data_width, writer_depth, reader_depth):
+    def __init__(self, data_width, writer_depth, reader_depth, dynamic_depth=True):
         self.sink        = stream.Endpoint(dma_layout(data_width))
         self.source      = stream.Endpoint(dma_layout(data_width))
 
         self.next_source = stream.Endpoint(dma_layout(data_width))
         self.next_sink   = stream.Endpoint(dma_layout(data_width))
 
-        self.reader_fifo_depth = CSRStorage(bits_for(reader_depth), reset=reader_depth,
-            description="DMA Reader FIFO depth (in {}-bit words).".format(data_width))
-        self.reader_fifo_level = CSRStatus(bits_for(reader_depth),
-            description="DMA Reader FIFO level (in {}-bit words).".format(data_width))
-        self.writer_fifo_depth = CSRStorage(bits_for(writer_depth), reset=writer_depth,
-            description="DMA Writer FIFO depth (in {}-bit words).".format(data_width))
-        self.writer_fifo_level = CSRStatus(bits_for(writer_depth),
-            description="DMA Writer FIFO level (in {}-bit words).".format(data_width))
+        # Reader FIFO Control/Status.
+        assert bits_for(reader_depth) < 24
+        self.reader_fifo_control = CSRStorage(fields=[
+            CSRField("depth", offset=0, size=bits_for(reader_depth), reset=reader_depth,
+                description="DMA Reader FIFO depth (in {}-bit words).".format(data_width)),
+            CSRField("scratch",    offset=24, size=4, description="Software Scratchpad."),
+            CSRField("level_mode", offset=31, values=[
+                ("``0b0``", "Report Instantaneous level."),
+                ("``0b1``", "Report `Minimal` level since last clear.")
+            ])
+        ])
+        self.reader_fifo_status = CSRStatus(fields=[
+            CSRField("level", offset=0, size=bits_for(reader_depth),
+                description="DMA Reader FIFO level (in {}-bit words).".format(data_width))
+            ])
+
+        # Writer FIFO Control/Status.
+        assert bits_for(writer_depth) < 24
+        self.writer_fifo_control = CSRStorage(fields=[
+            CSRField("depth", offset=0, size=bits_for(writer_depth), reset=writer_depth,
+                description="DMA Writer FIFO depth (in {}-bit words).".format(data_width)),
+            CSRField("scratch",    offset=24, size=4, description="Software Scratchpad."),
+            CSRField("level_mode", offset=31, values=[
+                ("``0b0``", "Report Instantaneous level."),
+                ("``0b1``", "Report `Maximal` level since last clear.")
+            ])
+        ])
+        self.writer_fifo_status = CSRStatus(fields=[
+            CSRField("level", offset=0, size=bits_for(reader_depth),
+                description="DMA Writer FIFO level (in {}-bit words).".format(data_width))
+            ])
 
         # # #
 
         depth_shift = log2_int(data_width//8)
 
+        # Reader FIFO.
         reader_fifo = SyncFIFO(dma_layout(data_width), reader_depth//(data_width//8), buffered=True)
-        writer_fifo = SyncFIFO(dma_layout(data_width), writer_depth//(data_width//8), buffered=True)
-        self.submodules += reader_fifo, writer_fifo
+        self.submodules += reader_fifo
         self.comb += [
+            # Connect Reader Sink to Reader FIFO when Level < Configured Depth.
             self.sink.connect(reader_fifo.sink, omit={"valid", "ready"}),
-            If(reader_fifo.level < self.reader_fifo_depth.storage[depth_shift:],
+            If((reader_fifo.level < self.reader_fifo_control.fields.depth[depth_shift:]) | (not dynamic_depth),
                 self.sink.connect(reader_fifo.sink, keep={"valid", "ready"})
             ),
+            # Connect Reader FIFO to Reader Source.
             reader_fifo.source.connect(self.next_source),
+        ]
+
+        # Store Min.
+        reader_fifo_level_min = Signal.like(reader_fifo.level)
+        self.sync += If(reader_fifo.level < reader_fifo_level_min, reader_fifo_level_min.eq(reader_fifo.level))
+        # Clear on Status write or when in Instantaneous mode.
+        reader_fifo_level_clr = (self.reader_fifo_status.re | (self.reader_fifo_control.fields.level_mode == 0))
+        self.sync += If(reader_fifo_level_clr, reader_fifo_level_min.eq(2**len(reader_fifo_level_min)-1))
+        # Return Reader FIFO level.
+        self.comb += [
+            # Instantaneous.
+            If(self.reader_fifo_control.fields.level_mode == 0,
+                self.reader_fifo_status.fields.level[depth_shift:].eq(reader_fifo.level)
+            # Min.
+            ).Else(
+                self.reader_fifo_status.fields.level[depth_shift:].eq(reader_fifo_level_min)
+            )
+        ]
+
+        # Writer FIFO.
+        writer_fifo = SyncFIFO(dma_layout(data_width), writer_depth//(data_width//8), buffered=True)
+        self.submodules += writer_fifo
+        self.comb += [
+            # Connect Writer Sink to Writer FIFO when Level < Configured Depth.
             self.next_sink.connect(writer_fifo.sink, omit={"valid", "ready"}),
-            If(writer_fifo.level < self.writer_fifo_depth.storage[depth_shift:],
+            If((writer_fifo.level < self.writer_fifo_control.fields.depth[depth_shift:]) | (not dynamic_depth),
                 self.next_sink.connect(writer_fifo.sink, keep={"valid", "ready"})
             ),
+            # Connect Writer FIFO to Writer Source.
             writer_fifo.source.connect(self.source),
-            self.reader_fifo_level.status[depth_shift:].eq(reader_fifo.level),
-            self.writer_fifo_level.status[depth_shift:].eq(writer_fifo.level),
+        ]
+        # Store Max.
+        writer_fifo_level_max = Signal.like(writer_fifo.level)
+        self.sync += If(writer_fifo.level > writer_fifo_level_max, writer_fifo_level_max.eq(writer_fifo.level))
+        # Clear on Status write or when in Instantaneous mode.
+        writer_fifo_level_clr = (self.writer_fifo_status.re | (self.writer_fifo_control.fields.level_mode == 0))
+        self.sync += If(writer_fifo_level_clr, writer_fifo_level_max.eq(0))
+        # Return Writer FIFO level.
+        self.comb += [
+            # Instantaneous.
+            If(self.writer_fifo_control.fields.level_mode == 0,
+                self.writer_fifo_status.fields.level[depth_shift:].eq(writer_fifo.level)
+            # Min.
+            ).Else(
+                self.writer_fifo_status.fields.level[depth_shift:].eq(writer_fifo_level_max)
+            )
         ]
 
 # LitePCIeDMA --------------------------------------------------------------------------------------
